use strict;
use warnings;
our $VERSION = '0.1';
our %IRSSI = (
    contact     => 'Nei @ anti@conference.jabber.teamidiot.de',
    url         => "http://anti.teamidiot.de/",
    name	=> 'teddy_cli',
    description => 'Module that conects to a WebSocket Proxy',
   );
use Irssi;
use Irssi::TextUI;

use File::Basename 'dirname';
use File::Spec;
use Cwd 'abs_path';
use constant ScriptFile => __FILE__;

use lib File::Spec->catdir(dirname(abs_path(+ScriptFile)), 'lib');

use IrssiX::include;
use Nei::Mojo::Reactor::Irssi4;
use Mojolicious::Lite;
use Mojo::UserAgent;
use Future;
use Variable::Disposition 'retain_future';

app->mode('production');
app->log->level('fatal');

use Encode;
use Scalar::Util qw(weaken looks_like_number);
use List::Util 'shuffle';
use Hash::Util 'lock_keys';
use POSIX 'strftime';
use Digest::SHA 'hmac_sha256_base64';
use Storable 'dclone';

use again 'Nei::Mojo::Transaction::WebSocketWithRawlogSend';

my $ua = Mojo::UserAgent->new;
my @connections;
my %lwin_map;
my %win_numbers;
my %lview_map;
my %open_lines;

sub shortdump {
    use Data::Dumper;
    local $Data::Dumper::Indent = 0;
    local $Data::Dumper::Quotekeys = 0;
    local $Data::Dumper::Sortkeys = 1;
    local $Data::Dumper::Useqq = 1;
    local $Data::Dumper::Terse = 1;
    local $Data::Dumper::Varname = '';
    local $Data::Dumper::Pair = ':';
    Dumper(+shift)
}

sub DEBUG (@) {
    my $win = Irssi::window_find_name('(ws)') or return;
    $win->print("@_");
    return;
}

sub send_x {
    my ($tx, $first) = @_;
    my @subq;
    for (sort keys %$first) {
	push @subq, my $q = Future->new;
	$tx->{tc_needed}{$_} = $q->done_cb;
    }
    $tx->send({json => $first});
    @subq == 1 ? pop @subq : Future->need_all(@subq)
}

sub send_p {
    my ($tx, $first) = @_;
    my $waits = $tx->{tc_waits} ||= +{};
    my $id;
    if (exists $first->{id}) {
	$id = $first->{id};
    }
    else {
	do {
	    $id = substr rand, 2;
	} while (exists $waits->{$id});
	$first->{id} = $id;
    }
    my $q = Future->new;
    $waits->{$id} = $q->done_cb;
    $tx->send({json => $first});
    $q
}

sub retain_f_for_cancel ($$) {
    my ($tx, $q) = @_;

    $q->on_fail(sub { die @_ });
    my $p = retain_future $q;

    weaken $p;
    push @{$tx->{open_q}}, $p;
}

sub _color_hack {
    my $id = shift;
    #"\\cC" . (join "\\cC", map { join ",", unpack "(A2)*", $_ } unpack "(A4)*", $id) . "\\cO"
    (join '', map { sprintf "\cD%02d", $_ } unpack "(A2)*", $id)."\cDg"
}

my $irssi_mumbo2 = qr/\cD[`-i]|\cD[&-@\xff].|\c_|\cV|\cD#..../;

sub _from_uni2 {
    my $var = shift;
    my $format;
    my $out = '';
    for my $chunk (split /($irssi_mumbo2)/, $var, -1) {
	if (defined $chunk) {
	    if ($format) {
		utf8::downgrade($chunk);
		$out .= $chunk;
	    } else {
		utf8::encode($chunk);
		$out .= $chunk
	    }
	}
	$format = !$format;
    }
    $out
}

sub irssi_line_removed {
    my ($view, $line, $prev_line) = @_; # TEXTBUFFER_VIEW_REC *view, LINE_REC *line, LINE_REC *prev_line
    if (my $win = $lview_map{ $view->{_irssi} }) {
	delete $win->{local_view_lines}{ $line->{_irssi} };
    }
    delete $open_lines{$view->{_irssi}}{$line->{_irssi}};
}

sub irssi_window_destroyed {
    my ($lwin) = @_;
    delete $lwin_map{ $lwin->{_irssi} };
    delete $lview_map{ $lwin->view->{_irssi} };
    delete $win_numbers{ $lwin->{_irssi} };
}

sub irssi_window_created {
    my ($lwin) = @_;
    $win_numbers{ $lwin->{_irssi} } = $lwin->{refnum};
}

sub irssi_window_refnum_changed {
    my ($lwin, $oldnumber) = @_;
    $win_numbers{ $lwin->{_irssi} } = $lwin->{refnum};
}

sub irssi_window_changed {
    my ($lwin, $loldwin) = @_;
    return unless $loldwin;
    if (my $win = $lwin_map{ $lwin->{_irssi} }) {
	sync_lines($win->{ct}{tx}, $win);
    }
}

sub irssi_page_scrolled {
    my ($lwin) = @_;
    if (my $win = $lwin_map{ $lwin->{_irssi} }) {
	if ($win->{earliest_line}{id} == ($win->{local_view_lines}{($lwin->view->{startline}//{_irssi=>0})->{_irssi}}//0)) {
	    load_more_lines($win->{ct}{tx}, $win);
	}
    }
}

sub irssi_send_line {
    local our $PREFER_REMOTE_COMMAND = 1;
    &Irssi::signal_continue;
}

sub _debug_artificial_delay {
    return sub{Future->done(@_)} unless our $DELAY_TESTING;
    my $delay = shift;
    sub {
	my (@args) = @_;
	my $q = Future->new;
	Irssi::timeout_add_once($delay*1000, sub {
				    $q->done(@args)
				}, '');
	$q;
    }
}

sub irssi_send_command {
    my ($line, $server, $item) = @_;
    return if our $LOCAL_COMMAND;
    my $lwin = $item ? $item->window : Irssi::active_win;
    my $cmdchars = join '|', map { quotemeta } split //, Irssi::parse_special('$K');
    if ($line =~ s/^($cmdchars)local\s//i) {
	local $LOCAL_COMMAND = 1;
	Irssi::signal_continue($line, $server, $item);
	return;
    } elsif (!$lwin_map{ $lwin->{_irssi} }) {
	return;
    }
    elsif ($line =~ /^($cmdchars)(scrollback|statusbar|window\s(last|up|down|goto|\d+)|go|ack)(\s|$)/i) {
	local $LOCAL_COMMAND = 1;
	Irssi::signal_continue($line, $server, $item);
	return;
    }
    elsif (my $win = $lwin_map{ $lwin->{_irssi} }) {
	unless ($line =~ s/^($cmdchars)remote\s//i) {
	    return unless our $PREFER_REMOTE_COMMAND;
	}
	Irssi::signal_stop;
	my $tx = $win->{ct}{tx};

	$lwin->print("%w... %K".$line=~s/%/%%/gr, MSGLEVEL_NEVER);
	my $bottom = $lwin->view->{bottom};
	my $loading_line = $lwin->view->{buffer}{cur_line};
	$open_lines{$lwin->view->{_irssi}}{$loading_line->{_irssi}} = 1;

	retain_f_for_cancel $tx,
	    send_p($tx, +{input=>{window=>$win->{id},data=>$line}})
	->then(_debug_artificial_delay(1))
	->then(sub {
		   my ($tx, $msg) = @_;
		   $lwin->view->remove_line($loading_line) if delete $open_lines{$lwin->view->{_irssi}}{$loading_line->{_irssi}};
		   $lwin->view->redraw;
		   $lwin->command('^scrollback end') if $bottom && !$lwin->view->{bottom};
		   Future->done
	       })
    }
}

use Devel::Confess;

sub load_more_lines {
    my ($tx, $win) = @_;
    my $lwin = _get_local_win($win);
    my $after = $lwin->view->{startline};
    $after = $after->prev if $after;
    $lwin->print_after($after, MSGLEVEL_NEVER, "%Kloading...", 1);
    my $loading_line = $lwin->last_line_insert;
    $open_lines{$lwin->view->{_irssi}}{$loading_line->{_irssi}} = 1;
    $lwin->view->redraw;
    $lwin->view->scroll_line($loading_line);
    my $scroll_after = $loading_line->next;
    #$lwin->command('^window scroll off');
    retain_f_for_cancel $tx,
#	_debug_artificial_delay(1)->()
	send_p($tx, +{line=>{get=>{
	    text=>1,
	    lv=>1,
	    $win->{view} => {before_line=>$win->{earliest_line}{id}},
	   }}})
    ->then(_debug_artificial_delay(1))
    ->then(sub {
	       my ($tx, $msg) = @_;
	       my $lwin = _get_local_win($win);
	       my $lview = $lwin->view;
	       my $start_line = $lwin->view->{startline};
	       ($lview) = _insert_lines_into_view($win, $lwin, $msg, $win->{local_view_lines}, $start_line, $loading_line);
	       $lview->remove_line($loading_line) if delete $open_lines{$lview->{_irssi}}{$loading_line->{_irssi}};
	       $lview->redraw;
	       #$lwin->command('^window scroll default');
	       my $scroll_to;
	       $scroll_to = $scroll_after->prev if $scroll_after;
	       $lwin->view->scroll_line($scroll_to) if $scroll_to;
	       Future->done
	   })
}

sub _insert_lines_into_view {
    my ($win, $lwin, $msg, $new_lines, $curline, $loading_line) = @_;

    my $data = $msg->{line}{get};
    my $lview = $lwin->view;
    $lwin->print('`', MSGLEVEL_NEVER);
    my $placeholder_line = $lwin->view->{buffer}{cur_line};
    $open_lines{$lwin->view->{_irssi}}{$placeholder_line->{_irssi}} = 1;
    for my $view (keys %$data) {
	next unless @{$data->{$view}};
	$win->{earliest_line} = $data->{$view}[0];
	while ($curline
		   && ($curline->{_irssi} == $loading_line->{_irssi}
			   || $curline->{_irssi} == $placeholder_line->{_irssi}
			   || $curline->{info}{time} > $data->{$view}[0]{time})) {
	    $curline = $curline->prev;
	}
	for my $line (@{$data->{$view}}) {
	    if (my $next = $curline) {
		while ($next
			   && ($next->{_irssi} == $loading_line->{_irssi}
				   || $next->{_irssi} == $placeholder_line->{_irssi}
				   || $next->{info}{time} < $line->{time})) {
		    $curline = $next;
		    $next = $next->next;
		}
	    }
	    #$lwin->gui_printtext_after($curline, MSGLEVEL_NEVER, _from_uni2($line->{text})."\n", $line->{time});
	    $lwin->print_after($curline, MSGLEVEL_NEVER, _from_uni2($line->{text})=~s/%/%%/gr, $line->{time});
	    $curline = $lwin->last_line_insert;
	    $new_lines->{ $curline->{_irssi} } = $line->{id};
	}
    }
    $lview->remove_line($placeholder_line) if delete $open_lines{$lview->{_irssi}}{$placeholder_line->{_irssi}};
    ($lview)
}

sub sync_lines {
    my ($tx, $win) = @_;
    my $lwin = _get_local_win($win);
    my $after = $lwin->view->{startline};
    $after = $after->prev if $after;
    $lwin->print_after($after, MSGLEVEL_NEVER, "%Kloading...", 1);
    my $loading_line = $lwin->last_line_insert;
    $lwin->view->redraw;
    $open_lines{$lwin->view->{_irssi}}{$loading_line->{_irssi}} = 1;
    retain_f_for_cancel $tx,
	send_p($tx, +{line=>{get=>{
	    text=>1,
	    lv=>1,
	    $win->{view} => {},
	   },
			     sub=>{add=>{view=>$win->{view},text=>1,lv=>1},
				   alter=>{view=>$win->{view},text=>1,lv=>1}}}})
    ->then(_debug_artificial_delay(1))
    ->then(sub {
	       my ($tx, $msg) = @_;
	       my $lwin = _get_local_win($win);
	       my $lview = $lwin->view;
	       $win->{local_view} = $lview->{_irssi};
	       $lview_map{ $lview->{_irssi} } = $win;
	       weaken $lview_map{ $lview->{_irssi} };
	       my $new_lines = +{};
	       my $start_line = $lwin->view->{buffer}{cur_line};
	       ($lview) = _insert_lines_into_view($win, $lwin, $msg, $new_lines, $start_line, $loading_line);
	       $lview->remove_line($loading_line) if delete $open_lines{$lview->{_irssi}}{$loading_line->{_irssi}};
	       for my $lp (keys %{$win->{local_view_lines}//+{}}) {
		   $lview->remove_line(bless +{_irssi=>$lp} => 'Irssi::TextUI::Line');
	       }
	       $win->{local_view_lines} = $new_lines;
	       $lview->redraw;
	       $tx->send({json => +{window=>{dehilight=>[$win->{id}]}}});
	       Future->done
	   });
}

sub _win_item_changed {
    my ($win, $active) = @_;
    my $lwin = _get_local_win($win) || return;
    my $ct = $win->{ct};

    if (@_ > 1) {
	$win->{active} = $active;
    }

    if ($win->{active}) {
	$lwin->set_name( $ct->{items}{ $win->{active} }{visible_name}
			     . _color_hack($win->{id}) . _color_hack(0+$ct) );
    } elsif ($win->{name}) {
	$lwin->set_name($win->{name} . _color_hack($win->{id}) . _color_hack(0+$ct) );
    } elsif (length $lwin->{name}) {
	$lwin->set_name('');
    }
}

sub _get_local_win {
    my ($win) = @_;
    Irssi::window_find_refnum( $win_numbers{ $win->{local} } )
}

sub _register_window_map {
    my ($win, $lwin, $ct, $awin) = @_;
    my $lid = $lwin->{_irssi};

    $ct->{windows}{ $win->{id} } = $win;

    $win->{local} = $lid;
    $win->{ct} = $ct;
    weaken $win->{ct};
    $lwin_map{ $lid } = $win;
    weaken $lwin_map{ $lid };
    unless ($awin && $awin->{_irssi} == $lid) {
	$lwin->activity($win->{data_level}, $win->{hilight_color});
    }
    unless ($win->{local_name}) { # not a merged
	$lwin->set_refnum($win->{refnum} + $ct->{win_offset});
	_win_item_changed($win);
    }
    $lwin;
}

sub _map_windows {
    my ($ct, @windows) = @_;

    my @local_wins = Irssi::windows;
    my %named_wins = map { ($_->{name} => $_) } grep { length $_->{name} } @local_wins;
    my $offset = $ct->{win_offset} // (sort { $b <=> $a } map { $_->{refnum} } @local_wins)[0];

    my $awin = Irssi::active_win;

    for my $win (@windows) {
	next unless exists $named_wins{$win->{name}};
	$win->{local_name} = 1;
	_register_window_map($win => $named_wins{$win->{name}}, $ct, $awin);
	$offset--;
	DEBUG "merge: ".$win->{name}."(".$named_wins{$win->{name}}{refnum}.")";
    }
    $ct->{win_offset} //= $offset;
    for my $win (@windows) {
	next if $win->{local_name};

	_register_window_map($win => Irssi::Windowitem::window_create(undef, 1),
			     $ct, $awin);
    }
}

sub _map_items {
    my ($ct, @items) = @_;
    for (@items) {
	$ct->{items}{ $_->{id} } = $_;
    }
    return;
}

my %dispatcher = (
    _login_done => sub {
	my ($tx) = @_;
	print CLIENTCRAP "Login succeeded";
	$tx->{logged_in} = 1;
	retain_f_for_cancel $tx,
	    send_p($tx, +{window=>{get=>{lv=>1},
			       sub=>{
				   ex=>{lv=>1},
				   attr=>{},
				   change=>{remote=>0},
				   act=>{},}},
			  item=>{get=>{lv=>1},
				 sub=>{
				     ex=>{lv=>1},
				     attr=>{},
				 }},
			  bind=>{beep=>"{}"},
		     })
        ->then(sub {
		   my ($tx, $msg) = @_;
		   DEBUG "got P response!!! tx=".$tx." msg=".$msg." [".(join ",", sort keys %$msg)."]";

		   my $ct = $tx->{ct};

		   _map_items($ct, @{$msg->{item}{get}});

		   _map_windows($ct, @{$msg->{window}{get}});

		   if (my $win = $lwin_map{ Irssi::active_win->{_irssi} }) {
		       sync_lines($tx, $win);
		   }
		   Future->done;
	       });
    },
    'window activity' => sub {
	my ($tx, $ev, $msg) = @_;
	my $win = $tx->{ct}{windows}{$msg->{id}};
	unless ($win && exists $win->{local}) {
	    DEBUG "unmapped window: ".$msg->{id};
	    next;
	}
	_get_local_win($win)
	    ->activity($msg->{data_level}, $msg->{hilight_color});
    },
    'line added' => sub {
	my ($tx, $ev, $msg) = @_;
	for my $view (keys %$msg) {
	    my ($win) = grep { $view == $_->{view} } values %{$tx->{ct}{windows}};
	    unless ($win && exists $win->{local}) {
		DEBUG "window not found: view $view";
		next;
	    }
	    my $lwin = _get_local_win($win);
	    for my $line (@{$msg->{$view}}) {
		$lwin->print(_from_uni2($line->{text})=~s/%/%%/gr, MSGLEVEL_NEVER);
		my $curline = $lwin->view->{buffer}{cur_line};
		$win->{local_view_lines}{ $curline->{_irssi} } = $line->{id};
	    }
	}
    },
    'line removed' => sub {
	my ($tx, $ev, $msg) = @_;
	for my $view (keys %$msg) {
	    my ($win) = grep { $view == $_->{view} } values %{$tx->{ct}{windows}};
	    unless ($win && exists $win->{local}) {
		DEBUG "window not found: view $view";
		next;
	    }
	    my %ids = map { ($_ => undef) } @{$msg->{$view}};
	    my $local_view_lines = $win->{local_view_lines} // +{};
	    my @lps = grep { exists $ids{ $local_view_lines->{$_} } } keys %$local_view_lines;
	    my $lwin = _get_local_win($win);
	    my $lview = $lwin->view;
	    my $bottom = $lview->{bottom};
	    my $delete;
	    for my $lp (@lps) {
		$lview->remove_line(bless +{_irssi=>$lp} => 'Irssi::TextUI::Line');
		$delete = 1;
	    }
	    $lview->redraw if $delete;
	    delete @{$local_view_lines}{@lps};
	    $lwin->command('^scrollback end') if $bottom && !$lwin->view->{bottom};
	}
    },
    'line added after' => sub {
	my ($tx, $ev, $msg) = @_;
	for my $view (keys %$msg) {
	    my ($win) = grep { $view == $_->{view} } values %{$tx->{ct}{windows}};
	    unless ($win && exists $win->{local}) {
		DEBUG "window not found: view $view";
		next;
	    }
	    my %ids = map { ($_->{prevline} => undef) } grep { defined $_->{prevline} } @{$msg->{$view}};
	    my $local_view_lines = $win->{local_view_lines} // +{};
	    my @lps = grep { exists $ids{ $local_view_lines->{$_} } } keys %$local_view_lines;
	    my %ils = map { ($local_view_lines->{$_} => $_) } @lps;
	    my $curline;
	    my $lwin = _get_local_win($win);
	    my $skip_block;
	    for my $line (@{$msg->{$view}}) {
		if (exists $line->{prevline}) {
		    $skip_block = undef;
		    if (!defined $line->{prevline}) {
			$curline = undef;
		    } else {
			unless (exists $ils{ $line->{prevline} }) {
			    $skip_block = 1;
			    next;
			}
			$curline = bless +{_irssi => $ils{ $line->{prevline} }} => 'Irssi::TextUI::Line';
		    }
		} elsif ($skip_block) {
		    next;
		}
		$lwin->print_after($curline, MSGLEVEL_NEVER, _from_uni2($line->{text})=~s/%/%%/gr, $line->{time});
		$curline = $lwin->last_line_insert;
		$win->{local_view_lines}{ $curline->{_irssi} } = $line->{id};
	    }
	}
    },
    'beep' => sub {
	Irssi::signal_emit(beep => ());
    },
    'window created' => sub {
	my ($tx, $ev, $msg) = @_;
	_map_windows($tx->{ct}, $msg);
    },
    'window item new' => sub {
	my ($tx, $ev, $msg) = @_;
	_map_items($tx->{ct}, $msg);
    },
    'window item changed' => sub {
	my ($tx, $ev, $msg) = @_;
	my $ct = $tx->{ct};
	_win_item_changed($ct->{windows}{ $msg->{id} }, $msg->{active})
	    if defined $ct->{windows}{ $msg->{id} };
    },
    'window changed' => sub {
	my ($tx, $ev, $msg) = @_;
	_get_local_win($tx->{ct}{windows}{ $msg->{id} })->set_active;
    },
    'window level changed' => sub {
	# ignored
    },
    'window destroyed' => sub {
	my ($tx, $ev, $msg) = @_;
	my $lwin = _get_local_win(delete $tx->{ct}{windows}{ $msg });
	$lwin->destroy unless $lwin->items;
    },
    'window item remove' => sub {
	my ($tx, $ev, $msg) = @_;
	my $it = delete $tx->{ct}{items}{ $msg };
	if (my $win = $tx->{ct}{windows}{ $it->{window} }) {
	    if (defined $win->{active} && $win->{active} == $msg) {
		_win_item_changed($win, undef);
	    }
	}
    },
   );

sub tc_dispatch {
    my ($tx, $fun, $msg) = @_;
    if (exists $dispatcher{$fun}) {
	$dispatcher{$fun}->($tx, $fun, $msg);
    }
    else {
	print CLIENTNOTICE "Unimplemented event $fun";
    }
}

sub ipw_rawlog_send {
    my ($tx, $msg) = @_;
    DEBUG "sent message: ".(shortdump($msg)=~s/%/%%/gr);
}

sub ws_message_handler {
    my ($tx, $msg) = @_;
    my $sd = shortdump($msg);
    DEBUG "WebSocket message: ".($sd=~s/%/%%/gr)
	unless $sd eq "{}";

    return unless 'HASH' eq ref $msg;

    # send_x
    if (exists $tx->{tc_needed}) {
	for my $fun (keys %$msg) {
	    (delete $tx->{tc_needed}{$fun})->($tx, delete $msg->{$fun})
		if $tx->{tc_needed}{$fun};
	}
	delete $tx->{tc_needed} unless %{$tx->{tc_needed}};
    }

    # send_p
    if (exists $msg->{id}) {
	my $id = delete $msg->{id};
	if (exists $tx->{tc_waits}{$id}) {
	    (delete $tx->{tc_waits}{$id})->($tx, $msg);
	    return;
	}
	print CLIENTCRAP "Message ID $id was not waited for";
    }

    # events
    for my $f (keys %$msg) {
	tc_dispatch($tx, $f, delete $msg->{$f});
    }
}

sub ws_disconnect_handler {
    my ($tx, $code, $reason) = @_;
    for my $cq (@{$tx->{open_q}//[]}) {
	next unless $cq;
	$cq->cancel;
    }
    print CLIENTCRAP "WebSocket closed with status $code.";
}

sub ws_login {
    my ($tx, $pass) = @_;

    my $secret = substr +(join '', map { chr } shuffle 0x21..0x7e), 0, 32;

    send_x($tx, +{ challenge => $secret })
    ->then(sub {
	       my ($tx, $challenge) = @_;

	       send_x($tx, { login => hmac_sha256_base64(
		   $pass, $challenge . $secret) });
	   })
    ->then(sub {
	       my ($tx, $login) = @_;
	       if ($login) {
		   tc_dispatch($tx, _login_done => ())
	       }
	       Future->done;
	   });
}

sub connect_teddy {
    my ($addr, $port, $pass) = @_;
    print "Connecting to: $addr:$port/teddy";
    my $conntrack = {
	addr => $addr,
	port => $port,
	pass => $pass,
    };
    $ua->websocket(
	"$addr:$port/teddy"
	=> {'Sec-WebSocket-Extensions' => 'permessage-deflate'}
	=> sub {
	    my ($ua, $tx) = @_;

	    $tx->{ct} = $conntrack;
	    weaken $tx->{ct};

	    # Check if WebSocket handshake was successful
	    (print CLIENTERROR 'WebSocket handshake failed!'), return
		unless $tx->is_websocket;

	    bless $tx => 'Nei::'.(ref $tx).'WithRawlogSend';
	    $tx->{rawlog_on} = \1;

	    $tx->on(finish => \&ws_disconnect_handler);
	    $tx->on(json => \&ws_message_handler);

	    # login
	    retain_f_for_cancel $tx,
		ws_login($tx, $pass);

	    $conntrack->{tx} = $tx;
	    weaken $conntrack->{tx};
	});
    push @connections, $conntrack;
}

sub init {
    for my $lwin (Irssi::windows) {
	$win_numbers{ $lwin->{_irssi} } = $lwin->{refnum};
    }
}

init();

Irssi::command_bind 'wsconnect' => sub {
    my ($data, $server, $witem) = @_;
    $data =~ s/\s+$//;
    my ($addr, $port, $pass) = split ' ', $data;
    connect_teddy($addr, $port, $pass);
};

Irssi::command_bind 'wssend' => sub {
    my ($data, $server, $witem) = @_;
    my ($id, $eval) = split ' ', $data, 2;
    local $@;
    my $code = defined $eval && do { eval $eval; };
    if (my $error = $@) {
	$error =~ s/ at .*//;
	print CLIENTERROR "Error in data: $error";
	return;
    }
    unless (ref $code) {
	print CLIENTERROR "Error in data: Not a reference";
	return;
    }
    if ($id > @connections || !$connections[$id]) {
	print CLIENTERROR "Invalid connection id";
	return;
    }
    $connections[$id]{tx}->send({json => $code});
};

{ package Irssi::Nick }

Irssi::signal_register({'key send_line' => [qw[string ulongptr Irssi::UI::Keyinfo]]});
Irssi::signal_register({'gui page scrolled' => [qw[Irssi::UI::Window]]});
Irssi::signal_add_last +{
    'gui page scrolled' => 'irssi_page_scrolled',
};
Irssi::signal_add_first +{
    'window created' => 'irssi_window_created',
    'window refnum changed' => 'irssi_window_refnum_changed',
    'gui textbuffer line removed' => 'irssi_line_removed',
    'window destroyed' => 'irssi_window_destroyed',
    'send command' => 'irssi_send_command',
    'key send_line' => 'irssi_send_line',
};
Irssi::signal_add +{
    'window changed' => 'irssi_window_changed',
};


sub UNLOAD {
    $ua = undef;
    for my $ct (@connections) {
	$ct->{tx}->finish if $ct && $ct->{tx};
    }
    for my $view (keys %lview_map) {
	if ( $lview_map{$view} ) {
	    my $lview = bless +{_irssi=>$view} => 'Irssi::TextUI::TextBufferView';
	    my $removed;
	    for my $lp (keys %{ $lview_map{$view}{local_view_lines} // +{} }) {
		delete $lview_map{$view}{local_view_lines}{$lp};
		$lview->remove_line(bless +{_irssi=>$lp} => 'Irssi::TextUI::Line');
		$removed = 1;
	    }
	    $lview->redraw if $removed;
	}
	delete $lview_map{$view};
    }
    @connections=();
    undef;
}
